# Takes a set of SCI0 resource file(s) & recreates the map file (typically named RESOURCE.MAP)

# Preparation:
#    Place the RESOURCE.XXX files in the same directory as this script

# Usage:    perl recreateMapFile.pl resource.map 


# From the FreeSCI documentation:
#
# SCI0 resource.map
#
# The SCI0 map file format is pretty simple: It consists of 6-byte entries, terminated by the sequence
# 0xffff ffff ffff. The first 2 bytes, interpreted as little endian 16 bit integer, encode resource type
# (high 5 bits) and number (low 11 bits). The next 4 bytes are a 32 bit LE integer that contains the 
# resource file number in the high 6 bits, and the absolute offset within the file in the low 26 bits. 
# SCI0 performs a linear search to find the resource; however, multiple entries may match the search, 
# since resources may be present more than once (the inverse mapping is not injective).
#
# SCI0 resource.<nr>
#
# SCI0 resource entries start with a four-tuple of little endian 16 bit words, which we will call 
# (id, comp_size, decomp_size, method). id has the usual SCI0 semantics (high 5 are the resource type, 
# low 11 are its number). comp_size and decomp_size are the size of the compressed and the decompressed
# resource, respectively. The compressed size actually starts counting at the record position of 
# decomp_size, so it counts four bytes in addition to the actual content. method, finally, is the 
# compression method used to store the data.


# Do whatever you want with this file, except remove my name.
# Jeremiah Nellis, nellisjp@hotmail.com
# Version 1.1 - Jan 22nd, 2012


# MAP:
# Resource file #s
#    - Byte 6, left 6 bits, read right-to-left
#      Examples:  00000100 (hex 04) = resource.001 
#                 00010100 (hex 14) = resource.005

# Resource position in file
#    - Byte 3, 4 & 5
#      Calculation = convert each byte to decimal, then:
#          byte 3 + byte 4 x 256 + byte 5 * 65536
#      Example: 31 0C 02 = 49 + (12 * 256) + (2 * 65536) =  134193 (absolute position in file)
#               

use strict;
use warnings;

sub addToMap($$);
sub writeMap($);

if (@ARGV < 1) { 
      print "Usage: recreateMapFile.pl %1\n";
      print "  %1: output map file name\n";
      exit (1);
}

if (-e $ARGV[0]) {
  print "Specified file to write to (" . $ARGV[0] . ") already exists!  Aborting.\n";
  exit;
} 

my $outputMapFile = $ARGV[0];
my @mapFile;

# Loop through each resource file
my $filename = "resource";
for my $file ( <$filename.*> ) {
    my $resourceNumber = substr($file,index($file,'.') + 1);
	if(length($file) != 12 || !($resourceNumber =~ m/[0-9][0-9][0-9]/) ) {
	   print "Skipping " . $file . "\n";
	} else {
	   print "Processing $file\n";
	   addToMap($file, int($resourceNumber));
	}
}

writeMap($outputMapFile);
exit;

sub addToMap($$) {
    my($inputResFile, $resourceNumber) = @_;

	my $data = do {
	    local $/ = undef;
	    open my $fh, $inputResFile
		or die "could not open $inputResFile: $!";
	    binmode $fh;
	    <$fh>;
	};

	my $rest = $data;
	my $filePos = int(0);

	while ($rest) {
	
	   # get the header (first 8 characters)
	   my $hexVal = substr($rest,0,8);
	   
	   # length is in characters 2, 3 and 4
	   my $resLength = int(ord(substr($hexVal,2,1))) 
	                 + int(ord(substr($hexVal,3,1))) * 256
      			     - 5;
	
	   my $mapValue;
	   	   
	   # Get the header & the body for this header
	   my $hexBody = substr($rest, 8, $resLength);
	 
    

	   if(length($rest) >= $resLength + 9) {
          $rest = substr($rest, 9 + $resLength);
	   } else {
	      $rest = undef;
	   }

	   my $hexBodyLen = sprintf("%x", $filePos);

	   # left pad uneven hex values
	   if (length($hexBodyLen)%2==1) {
		   $hexBodyLen = "0" . $hexBodyLen;
	   }

   
	   # reverse the hex values and right pad with extra 00 if necessary
	   if(length($hexBodyLen) == 6) {
		   $hexBodyLen = substr($hexBodyLen,4,2) . substr($hexBodyLen,2,2) . substr($hexBodyLen,0,2);
	   }

	   if(length($hexBodyLen) == 4) {
		   $hexBodyLen = substr($hexBodyLen,2,2) . substr($hexBodyLen,0,2);
		   $hexBodyLen = $hexBodyLen . "00";
	   } 

	   if(length($hexBodyLen) == 2) {
		   $hexBodyLen = $hexBodyLen . "0000";
	   }


	   # write out the resource number (RESOURCE.XXX) and type (characters 1 and 2)
	   $mapValue = substr($hexVal,0,2);

	   # write out the length
	   $mapValue = $mapValue . pack('H*',$hexBodyLen);


	   # convert the resource number to binary, then to hex
	   my $resNum = unpack("B*", pack("N", $resourceNumber));
	   $resNum = substr($resNum,length($resNum) - 6) . "00";
       my $hex =  sprintf('%X', oct("0b$resNum"));
      	   
	   if(length($hex) == 1) {
	      $hex = "0" . $hex;
	   }
 
	   $mapValue = $mapValue . pack('H*', $hex);

	   push(@mapFile, $mapValue);
	  
	   # calculate the absolute position for the next resource
	   $filePos = $filePos + length($hexBody) + 9;

    }
}

sub writeMap($) {
    my($outputMapFile) = @_;
	my $mapEntry;
	
	print "Creating new map file (" . $outputMapFile . ") \n";

	open(RESOURCEMAP, ">" . $outputMapFile) || die;
	binmode RESOURCEMAP;

    foreach $mapEntry (@mapFile) {    
       print RESOURCEMAP $mapEntry;
	}   

	@mapFile = ();
	
	# Write out the file footer
	print RESOURCEMAP pack('H*', 'FFFFFFFFFFFF');
	
	close(RESOURCEMAP);
}